/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <bsd/string.h>
#include <unistd.h>

#include <doca_argp.h>
#include <doca_log.h>

#include "dpi_common.h"

DOCA_LOG_REGISTER(DPI_COMMON);

/*
 * ARGP Callback - Handle signatures file parameter
 *
 * @param [in]: Input parameter
 * @config [in/out]: Program configuration context
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
sig_file_callback(void *param, void *config)
{
	struct dpi_scan_config *conf = (struct dpi_scan_config *)config;
	const char *sig_file = (char *)param;

	if (!conf->is_grpc_mode) {
		if (access(sig_file, F_OK | R_OK) != 0) {
			DOCA_LOG_ERR("Failed to find file path pointed by: %s", sig_file);
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (strnlen(sig_file, MAX_FILE_PATH_SIZE) == MAX_FILE_PATH_SIZE) {
			DOCA_LOG_ERR("Entered file path %s exceeded buffer size of: %d", sig_file, MAX_FILE_PATH_SIZE - 1);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}

	strlcpy(conf->sig_file_path, sig_file, MAX_FILE_PATH_SIZE);

	return DOCA_SUCCESS;
}

doca_error_t
register_dpi_scan_params()
{
	doca_error_t result;
	struct doca_argp_param *sig_file_param;

	/* Create and register signatures file param */
	result = doca_argp_param_create(&sig_file_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create ARGP param: %s", doca_get_error_string(result));
		return result;
	}
	doca_argp_param_set_short_name(sig_file_param, "s");
	doca_argp_param_set_long_name(sig_file_param, "sig-file");
	doca_argp_param_set_description(sig_file_param, "Signatures file path on DPU");
	doca_argp_param_set_callback(sig_file_param, sig_file_callback);
	doca_argp_param_set_type(sig_file_param, DOCA_ARGP_TYPE_STRING);
	doca_argp_param_set_mandatory(sig_file_param);
	result = doca_argp_register_param(sig_file_param);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register program param: %s", doca_get_error_string(result));
		return result;
	}

	return DOCA_SUCCESS;
}
